package com.bergerkiller.bukkit.nolagg;

import java.io.DataInput;
import java.io.DataInputStream;
import java.io.DataOutput;
import java.io.DataOutputStream;
import java.io.File;
import java.lang.reflect.Field;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.logging.Level;

import org.bukkit.Bukkit;
import org.bukkit.craftbukkit.CraftWorld;

import net.minecraft.server.*;

@SuppressWarnings("rawtypes")
public class BufferedChunkRegionLoader extends ChunkRegionLoader implements IChunkLoader {

	public static void swapAll() {
		if (!NoLagg.useBufferedChunkLoading) return;
		for (org.bukkit.World world : Bukkit.getServer().getWorlds()) {
			swap(world);
		}
	}
	public static void swap(org.bukkit.World world) {
		swap(((CraftWorld) world).getHandle());
	}
	public static void swap(WorldServer world) {
		swap(world.chunkProviderServer);
	}
	public static void swap(ChunkProviderServer cps) {
		if (!NoLagg.useBufferedChunkLoading) return;
		try {
			Field e = ChunkProviderServer.class.getDeclaredField("e");
			e.setAccessible(true);
			IChunkLoader loader = (IChunkLoader) e.get(cps);
			//get needed fields
			Field a = ChunkRegionLoader.class.getDeclaredField("a");
			Field b = ChunkRegionLoader.class.getDeclaredField("b");
			Field c = ChunkRegionLoader.class.getDeclaredField("c");
			Field d = ChunkRegionLoader.class.getDeclaredField("d");
			a.setAccessible(true);
			b.setAccessible(true);
			c.setAccessible(true);
			d.setAccessible(true);
			
			//get old loader variables
			List vara = (List) a.get(loader);
			Set varb = (Set) b.get(loader);
			Object varc = (Object) c.get(loader);
			File vard = (File) d.get(loader);
			
			//clear old loader's data
			a.set(loader, new ArrayList());
			b.set(loader, new HashSet());
			c.set(loader, new Object());
			
			//create a new loader and set it's native members
			loader = new BufferedChunkRegionLoader(cps.world, vara, varb, varc, vard);
			a.set(loader, vara);
			b.set(loader, varb);
			c.set(loader, varc);
			d.set(loader, vard);
			

			
			e.set(cps, loader);
		} catch (Throwable t) {
			NoLagg.log(Level.SEVERE, "Failed to initialize the buffered chunk loader:");
			t.printStackTrace();
		}
	}

	private final WorldServer worldOwner;
	private List savequeue = new ArrayList();
	private Set pendingsave = new HashSet();
	private Object lock = new Object();
	private final File sourcefolder;

	private BufferedChunkRegionLoader(final WorldServer world, List savequeue, Set pendingsave, Object lock, File sourcefolder) {
		super(sourcefolder);
		this.savequeue = savequeue;
		this.pendingsave = pendingsave;
		this.lock = lock;
		this.sourcefolder = sourcefolder;
		this.worldOwner = world;
	}

	private static Field pendinga;
	private static Field pendingb;
	public static ChunkCoordIntPair getPendingA(Object from) {
		try {
			if (pendinga == null) {
				pendinga = from.getClass().getDeclaredField("a");
				pendinga.setAccessible(true);
			}
			return (ChunkCoordIntPair) pendinga.get(from);
		} catch (Throwable t) {
			t.printStackTrace();
			return null;
		}
	}
	public static NBTTagCompound getPendingB(Object from) {
		try {
			if (pendingb == null) {
				pendingb = from.getClass().getDeclaredField("b");
				pendingb.setAccessible(true);
			}
			return (NBTTagCompound) pendingb.get(from);
		} catch (Throwable t) {
			t.printStackTrace();
			return null;
		}
	}

	@Override
    public boolean c() {
        Object pendingchunktosave = null;
        ChunkCoordIntPair coord;
        synchronized (this.lock) {
            if (this.savequeue.size() <= 0) {
                return false;
            }

            pendingchunktosave = this.savequeue.remove(0);
            coord = getPendingA(pendingchunktosave);
            this.pendingsave.remove(coord);
        }
        
        if (pendingchunktosave != null) {
            try {
                DataOutputStream dataoutputstream = RegionFileCache.c(this.sourcefolder, coord.x, coord.z);
                NBTCompressedStreamTools.a(getPendingB(pendingchunktosave), (DataOutput) dataoutputstream);
                BufferedChunk chunk = BufferedChunk.removeUnloaded(this.worldOwner, coord.x, coord.z);
                if (chunk != null) {
                	dataoutputstream.flush();
        			chunk.setFree();
                }
                dataoutputstream.close();  
            } catch (Exception exception) {
                exception.printStackTrace();
            }
        }
        return true;
    }
	
	/*
	 * This is called when a chunk has to be loaded from file
	 */
	@Override
	public Chunk a(World world, int i, int j) {
		if (NoLagg.useBufferedChunkLoading) {
			try {

				NBTTagCompound nbttagcompound = null;
				ChunkCoordIntPair chunkcoordintpair = new ChunkCoordIntPair(i, j);
				synchronized (this.lock) {
					if (this.pendingsave.contains(chunkcoordintpair)) {
						Object obj;
						for (int k = 0; k < this.savequeue.size(); ++k) {
							obj = this.savequeue.get(k);
							if (getPendingA(obj).equals(chunkcoordintpair)) {
								nbttagcompound = getPendingB(obj);
								break;
							}
						}
					}
				}

				if (nbttagcompound == null) {
					DataInputStream datainputstream = RegionFileCache.b(this.sourcefolder, i, j);
					if (datainputstream == null) return null;

					//custom memory-efficient loading system
					BufferedChunk chunk = new BufferedChunk(world, i, j);
					try {
						chunk.load(datainputstream);
						return chunk;
					} catch (Exception ex) {
						chunk.setFree();
						NoLagg.log(Level.SEVERE, "An exception occurred while loading chunk " + chunk.toString() + ":");
						ex.printStackTrace();
						NoLagg.log(Level.WARNING, "Native methods will now try to load it...");				
					}

					//failure, wrong coords? Use native coding instead
					datainputstream = RegionFileCache.b(this.sourcefolder, i, j);
					if (datainputstream == null) return null;
					nbttagcompound = NBTCompressedStreamTools.a((DataInput) datainputstream);
				}

				if (!nbttagcompound.hasKey("Level")) {
					System.out.println("Chunk file at " + i + "," + j + " is missing level data, skipping");
					return null;
				} else if (!nbttagcompound.getCompound("Level").hasKey("Blocks")) {
					System.out.println("Chunk file at " + i + "," + j + " is missing block data, skipping");
					return null;
				} else {
					Chunk chunk = ChunkLoader.a(world, nbttagcompound.getCompound("Level"));

					if (!chunk.a(i, j)) {
						System.out.println("Chunk file at " + i + "," + j + " is in the wrong location; relocating. (Expected " + i + ", " + j + ", got " + chunk.x + ", " + chunk.z + ")");
						nbttagcompound.setInt("xPos", i);
						nbttagcompound.setInt("zPos", j);
						chunk = ChunkLoader.a(world, nbttagcompound.getCompound("Level"));
					}

					chunk.h();
					return chunk;
				}

			} catch (Throwable t) {
				NoLagg.log(Level.SEVERE, "Buffered chunk loader faced an incompatibility issue with the current build:");
				t.printStackTrace();
				NoLagg.log(Level.INFO, "Buffered chunk loading has been disabled until a server reload or restart is performed.");
				NoLagg.useBufferedChunkLoading = false;
			}
		}
		return super.a(world, i, j);
	}

}
